/********************************************************************************
    iframeenc - encode a single PPM image as MPEG I-Frame
 ********************************************************************************
    LibVideoCoding - video coding library
    Copyright (C) 2002  Dirk Farin

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ********************************************************************************/

#include <libvideocoding/fileio/inputstream_istr.hh>
#include <libvideocoding/mpeg2/syntax/pessource.hh>
#include <libvideocoding/mpeg2/syntax/demux.hh>
#include <libvideocoding/mpeg2/syntax/sliceconf.hh>
#include <libvideocoding/mpeg2/syntax/video.hh>
#include <libvideocoding/mpeg2/syntax/macroblock.hh>
#include <libvideocoding/mpeg2/decoder/video.hh>
#include <libvideocoding/mpeg2/encoder/quantize.hh>
#include <libvideocoding/dct/dct.hh>
#include <libvideogfx.hh>

#include <unistd.h>
#include <stdio.h>
#include <iostream>
#include <iomanip>
#include <fstream>
using namespace std;
using namespace videocoding;

int main(int argc,char** argv)
{
  try
    {
      if (argc != 2)
	{
	  cerr << "Usage: iframeenc image.ppm\nPipe output into file.mpg.\n";
	  return 5;
	}

      Image<Pixel> inputimg;
      ReadImage_PPM(inputimg,argv[1]);

      Image<Pixel> img;
      ChangeColorspace(img, inputimg, Colorspace_YUV, Chroma_420);


      int q = 4;   // quality: qscale ( 1 best / 31 worst )

      MP2_SequenceHeader seqhdr;
      seqhdr.SetDefaults(true);
      seqhdr.width  = img.AskWidth();
      seqhdr.height = img.AskHeight();
      seqhdr.framerate100 = 2500;

      MP2_GOPHeader gophdr;

      MP2_PictureHeader pichdr;
      pichdr.seqhdr = &seqhdr;
      pichdr.temporal_reference = 0;
      pichdr.picture_coding_type = MP2_PicType_I;
      pichdr.vbv_delay = 0;
      pichdr.has_coding_extension = seqhdr.IsMPEG2;
  
      BitBuffer bb;

      seqhdr.Write(bb);
      gophdr.Write(bb);
      pichdr.Write(bb);

      MP2_MacroblockEncoder mbenc;
      MP2_SliceConfiguration_RowSlices sliceconf;
      sliceconf.SetSizeInMBs(seqhdr.AskWidthInMBs(),seqhdr.AskHeightInMBs());

      MP2_SliceHeader slicehdr;
      slicehdr.pichdr = &pichdr;

      MP2_Quantizer* quant = GetQuantizer();
      quant->SetMPEG2(seqhdr.IsMPEG2);

      FDCTBlk8x8<Pixel>* fdct = GetFDCT8x8_8();

      const Pixel*const* yp = img.AskFrameY();
      const Pixel*const* up = img.AskFrameU();
      const Pixel*const* vp = img.AskFrameV();
      int ystride = img.AskBitmapY().AskStride();
      int cstride = img.AskBitmapU().AskStride();

      mbenc.SetOutputStream(bb);
      mbenc.SetSliceConfiguration(&sliceconf);

      for (int y=0;y<seqhdr.AskHeightInMBs();y++)
	for (int x=0;x<seqhdr.AskWidthInMBs();x++)
	  {
	    if (sliceconf.IsFirstInSlice(y,x))
	      {
		slicehdr.slice_vertical_position = y;
		slicehdr.Write(bb);
	      }

	    MP2_Macroblock mb;

	    mb.slicehdr = &slicehdr;
	    mb.skipped = false;
	    mb.use_field_DCT = false;
	    mb.motion.prediction_mode = MP2_Intra;

	    mb.qcode = q;

	    DCTBlock_8x8 blk;

	    fdct->CalcFDCT(blk, &yp[y*16  ][x*16  ], ystride);
	    quant->Quantize_Intra(mb.dct_y [0][0], blk, q*2);

	    fdct->CalcFDCT(blk, &yp[y*16  ][x*16+8], ystride);
	    quant->Quantize_Intra(mb.dct_y [0][1], blk, q*2);

	    fdct->CalcFDCT(blk, &yp[y*16+8][x*16  ], ystride);
	    quant->Quantize_Intra(mb.dct_y [1][0], blk, q*2);

	    fdct->CalcFDCT(blk, &yp[y*16+8][x*16+8], ystride);
	    quant->Quantize_Intra(mb.dct_y [1][1], blk, q*2);

	    fdct->CalcFDCT(blk, &up[y*8   ][x*8   ], cstride);
	    quant->Quantize_Intra(mb.dct_cb[0][0], blk, q*2);

	    fdct->CalcFDCT(blk, &vp[y*8   ][x*8   ], cstride);
	    quant->Quantize_Intra(mb.dct_cr[0][0], blk, q*2);

	    mbenc.Write(mb,y,x);
	  }

      bb.AlignToByte0();
      bb.WriteBits(1,24);
      bb.WriteBits(MP2_StartCode_ISO11172_End,8);
      bb.Flush();

#if 1
      uint8* data = bb.AskBuffer();
      for (int i=0;i<bb.AskLengthInBytes();i++)
	cout << data[i];
#endif
    }
  catch (const Excpt_Base& e)
    {
      MessageDisplay::Show(e);
    }

  return 0;
}
