/********************************************************************************
    LibVideoCoding - video coding library
    Copyright (C) 2002  Dirk Farin

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ********************************************************************************/

#include "libvideocoding/mpeg2/decoder/video.hh"
#include "libvideocoding/mpeg2/syntax/errors.hh"

#include <libvideogfx/arch/mmx.h>

#include <unistd.h>


namespace videocoding {
  using namespace videogfx;

  // current picture memory

  const int currentpic = 0;
  const int nextpic    = 1;
  const int lastpic    = 2;
  const int invalidpic = -1;


  MP2_VideoDecoder::MP2_VideoDecoder()
    : d_motioncomp(false)
  {
    d_idct = GetIDCT8x8_8();

    Reset();
  }


  MP2_VideoDecoder::~MP2_VideoDecoder()
  {
    delete d_idct;
  }


  void MP2_VideoDecoder::Reset()
  {
    d_sink = NULL;

    d_seqhdr = NULL;

    d_have_top = d_have_bottom = false;
    d_pic_mem = invalidpic;
    d_field_idx = 0;

    d_need_picdata = false;

    for (int i=0;i<2;i++)
      for (int k=0;k<2;k++)
	{
	  d_data_valid[k][i] = false;
	  d_picdata[k][i].Release();
	}
  }


  MP2_Macroblock* MP2_VideoDecoder::RequestMacroblock(int mbx,int mby)
  {
    AssertDescr(d_seqhdr,"macroblock requested before sequence header has been read");

    if (d_need_picdata)
      return &(d_picdata[d_pic_mem][d_field_idx].AskFrame()[mby][mbx]);
    else
      return &mbmem;
  }


  void MP2_VideoDecoder::SendSequenceHeader(const MP2_SequenceHeader& s)
  {
    bool init= (d_seqhdr==NULL);

    d_seqhdr = &s;

    if (init)
      {
	ImageParam spec;
	spec.width  = d_seqhdr->width;
	spec.height = d_seqhdr->height;
	spec.halign = spec.valign = 16;
	spec.border = 0;
	spec.colorspace = Colorspace_YUV;

	switch (d_seqhdr->chroma_format)
	  {
	  case MP2_Chroma_420: spec.chroma = Chroma_420; break;
	  case MP2_Chroma_422: spec.chroma = Chroma_422; break;
	  case MP2_Chroma_444: spec.chroma = Chroma_444; break;
	  }

	for (int i=0;i<3;i++)
	  d_img[i].Create(spec);
      }
  }


  inline void SwapImgs(Image<Pixel>& a,Image<Pixel>& b) { Image<Pixel> t=a; a=b; b=t; }


  void MP2_VideoDecoder::FlushFrame(int mem)
  {
    Assert(mem == currentpic || mem == nextpic);

    d_sink->SendImage(d_img[mem],
		      (d_data_valid[mem][0]) ? &d_pichdr[mem][0] : NULL,
		      (d_data_valid[mem][1]) ? &d_pichdr[mem][1] : NULL,
		      (d_data_valid[mem][0] && d_need_picdata) ? &d_picdata[mem][0] : NULL,
		      (d_data_valid[mem][1] && d_need_picdata) ? &d_picdata[mem][1] : NULL);

    d_data_valid[mem][0] = d_data_valid[mem][1] = false;
  }


  void MP2_VideoDecoder::FlushField(int mem,bool top)
  {
    FlushFrame(mem);
  }


  void MP2_VideoDecoder::SendPictureHeader(const MP2_PictureHeader& p)
  {
    if (!d_seqhdr)
      return;

    // --- flush current picture if a new frame begins

    bool begin_new_pic = false;

    switch (p.picture_structure)
      {
      case MP2_PicStruct_FramePicture:
	if (d_have_top || d_have_bottom)
	  begin_new_pic = true;
	break;

      case MP2_PicStruct_TopField:
	if (d_have_top)
	  begin_new_pic = true;
	break;

      case MP2_PicStruct_BottomField:
	if (d_have_bottom)
	  begin_new_pic = true;
	break;
      }

    if (!d_have_top && !d_have_bottom)
      begin_new_pic = true;

    if (begin_new_pic)
      {
	if (p.picture_coding_type == MP2_PicType_B)
	  {
	    if (d_data_valid[currentpic][0]) FlushFrame(currentpic);
	  }
	else
	  {
	    if (d_data_valid[currentpic][0]) FlushFrame(currentpic);
	    if (d_data_valid[nextpic][0])    FlushFrame(nextpic);

	    SwapImgs(d_img[nextpic], d_img[lastpic]);
	  }

	d_have_top = d_have_bottom = false;
      }


    // --- prepare decoding of picture

    if (begin_new_pic)
      {
	d_need_picdata = d_sink->NeedPictureData();

	if (d_need_picdata)
	  for (int i=0;i<2;i++)
	    for (int k=0;k<2;k++)
	      d_picdata[i][k].Create(d_seqhdr->AskWidthInMBs(),d_seqhdr->AskHeightInMBs());
      }

    if (begin_new_pic)
      {
	switch (p.picture_coding_type)
	  {
	  case MP2_PicType_I:
	  case MP2_PicType_P:
	    d_pic_mem = nextpic;
	    break;

	  case MP2_PicType_B:
	    d_pic_mem = currentpic;
	    break;
	  }

	d_field_idx = 0;
      }
    else
      {
	d_field_idx = 1;
      }


    if (p.picture_structure == MP2_PicStruct_FramePicture)
      {
	switch (p.picture_coding_type)
	  {
	  case MP2_PicType_I:
	    break;

	  case MP2_PicType_P:
	  case MP2_PicType_B:
	    d_motioncomp.InitializeReferences(&d_img[lastpic], &d_img[nextpic], NULL,
					      true,
					      false);
	    break;
	  }
      }
    else
      {
	d_motioncomp.InitializeReferences(&d_img[lastpic], &d_img[nextpic], &d_img[d_pic_mem],
					  p.picture_coding_type == MP2_PicType_P,
					  true,
					  d_field_idx == 0,
					  p.picture_structure == MP2_PicStruct_TopField);
      }


    switch (p.picture_structure)
      {
      case MP2_PicStruct_FramePicture:
	d_have_top = d_have_bottom = true;
	break;

      case MP2_PicStruct_TopField:
	d_have_top = true;
	break;

      case MP2_PicStruct_BottomField:
	d_have_bottom = true;
	break;
      }

    d_pichdr    [d_pic_mem][d_field_idx] = p;
    d_data_valid[d_pic_mem][d_field_idx] = true;
  }


  void MP2_VideoDecoder::SendMacroblock(MP2_Macroblock* mb,int mbx,int mby)
  {
    if (!d_seqhdr || !d_pichdr)   // TODO: d_pichdr ?
      return;

#if 0
    static int cnt=1;
    if (mbx==0 && mby==0)
      {
	cout << "PIC: " << cnt << endl;
	cnt++;
      }
#endif

    int lumx = mbx*16;
    int lumy = mby*16;
    int chrx = mbx* 8;  ///TMP
    int chry = mby* 8;  ///TMP

    Assert(d_pic_mem != invalidpic);

    Image<Pixel>& img = d_img[d_pic_mem];

    Bitmap<Pixel> ybm;
    Bitmap<Pixel> cbbm;
    Bitmap<Pixel> crbm;

    switch(d_pichdr[d_pic_mem][d_field_idx].picture_structure)
      {
      case MP2_PicStruct_FramePicture:
	ybm  = img.AskBitmapY();
	cbbm = img.AskBitmapCb();
	crbm = img.AskBitmapCr();
	break;

      case MP2_PicStruct_TopField:
	ybm  = img.AskBitmapY().CreateFieldView(true);
	cbbm = img.AskBitmapCb().CreateFieldView(true);
	crbm = img.AskBitmapCr().CreateFieldView(true);
	break;

      case MP2_PicStruct_BottomField:
	ybm  = img.AskBitmapY().CreateFieldView(false);
	cbbm = img.AskBitmapCb().CreateFieldView(false);
	crbm = img.AskBitmapCr().CreateFieldView(false);
	break;
      }

    Pixel*const* yp  = ybm.AskFrame();
    Pixel*const* cbp = cbbm.AskFrame();
    Pixel*const* crp = crbm.AskFrame();

    int ystride  = ybm.AskStride();
    int cbstride = cbbm.AskStride();
    int crstride = crbm.AskStride();


    // do motion compensation

#if 0
    cout << "MOTION " << mbx << "/" << mby << ": ";
    switch (mb->motion.prediction_mode)
      {
      case MP2_Intra:
	mb->motion.vector[0][0].h = 
	  mb->motion.vector[0][0].v = 
	  mb->motion.vector[0][1].h = 
	  mb->motion.vector[0][1].v = 0;
	cout << "IN ";
	break;

      case MP2_Frame_FrameBased:
	cout << "FR ";
	break;

      case MP2_Frame_FieldBased:
	cout << "FI ";
	break;
      }

    cout << (mb->motion.vector[0][0].h>>15) << ";" << (mb->motion.vector[0][0].v>>15) << " ";
    cout << (mb->motion.vector[0][1].h>>15) << ";" << (mb->motion.vector[0][1].v>>15) << endl;
#endif

#if 0
    cout << "pos: " << mbx << "/" << mby << " -> " << lumx << "/" << lumy << endl;

    cout << "strides: " << ystride << " " << cbstride << " " << crstride << endl;

    //cout << "skipped: " << (mb->skipped ? "yes":"no") << endl;

    switch (mb->motion.prediction_mode)
      {
      case MP2_Frame_FrameBased: cout << "Frame-Frame\n"; break;
      case MP2_Frame_FieldBased: cout << "Frame-Field\n"; break;
      default: cout << "OTHER\n"; break;
      }

    if (mb->motion.prediction_mode != MP2_Intra)
      {
	emms();
	cout << "MV fwd: " << mb->motion.vector[0][0] << " " << mb->motion.vector[1][0];
	cout << " " << ((mb->motion.vertical_field_select[0][0]==MP2_Pred_TopField) ? "top":"bottom");
	cout << " " << ((mb->motion.vertical_field_select[0][1]==MP2_Pred_TopField) ? "top":"bottom") << endl;
	cout << "MV bkw: " << mb->motion.vector[0][1] << " " << mb->motion.vector[1][1];
	cout << " " << ((mb->motion.vertical_field_select[1][0]==MP2_Pred_TopField) ? "top":"bottom");
	cout << " " << ((mb->motion.vertical_field_select[1][1]==MP2_Pred_TopField) ? "top":"bottom") << endl;
      }
#endif

    //cout << "PIC: " << ((int*)&yp [lumy][lumx]) << " " << ((int*)&cbp[chry][chrx]) << " " << ((int*)&crp[chry][chrx]) << endl;

#if 1
    d_motioncomp.CreatePrediction(*mb, lumx, lumy,
				  &yp [lumy][lumx],ystride,
				  &cbp[chry][chrx],cbstride,
				  &crp[chry][chrx],crstride);
#endif

    // do IDCT

    //if (d_pichdr->picture_coding_type == MP2_PicType_I)
    {
      bool add = !(mb->motion.prediction_mode == MP2_Intra);

      if (mb->use_field_DCT)
	{
	  if (mb->dct_y[0][0].may_contain_data)
	    d_idct->CalcIDCT(&yp[lumy  ][lumx  ],ystride*2, mb->dct_y[0][0],add);
	  if (mb->dct_y[0][1].may_contain_data)
	    d_idct->CalcIDCT(&yp[lumy  ][lumx+8],ystride*2, mb->dct_y[0][1],add);
	  if (mb->dct_y[1][0].may_contain_data)
	    d_idct->CalcIDCT(&yp[lumy+1][lumx  ],ystride*2, mb->dct_y[1][0],add);
	  if (mb->dct_y[1][1].may_contain_data)
	    d_idct->CalcIDCT(&yp[lumy+1][lumx+8],ystride*2, mb->dct_y[1][1],add);
	  if (mb->dct_cb[0][0].may_contain_data)
	    d_idct->CalcIDCT(&cbp[chry][chrx],cbstride, mb->dct_cb[0][0],add);
	  if (mb->dct_cr[0][0].may_contain_data)
	    d_idct->CalcIDCT(&crp[chry][chrx],crstride, mb->dct_cr[0][0],add);
	}
      else
	{
	  if (mb->dct_y[0][0].may_contain_data)
	    d_idct->CalcIDCT(&yp[lumy  ][lumx  ],ystride, mb->dct_y[0][0],add);
	  if (mb->dct_y[0][1].may_contain_data)
	    d_idct->CalcIDCT(&yp[lumy  ][lumx+8],ystride, mb->dct_y[0][1],add);
	  if (mb->dct_y[1][0].may_contain_data)
	    d_idct->CalcIDCT(&yp[lumy+8][lumx  ],ystride, mb->dct_y[1][0],add);
	  if (mb->dct_y[1][1].may_contain_data)
	    d_idct->CalcIDCT(&yp[lumy+8][lumx+8],ystride, mb->dct_y[1][1],add);
	  if (mb->dct_cb[0][0].may_contain_data)
	    d_idct->CalcIDCT(&cbp[chry][chrx],cbstride, mb->dct_cb[0][0],add);
	  if (mb->dct_cr[0][0].may_contain_data)
	    d_idct->CalcIDCT(&crp[chry][chrx],crstride, mb->dct_cr[0][0],add);
	}
    }

#if 0
    if (mb->skipped)
      {
	DrawLine(img.AskBitmapY(), lumx   ,lumy,lumx+15,lumy+15, (Pixel)255);
	DrawLine(img.AskBitmapY(), lumx+15,lumy,lumx   ,lumy+15, (Pixel)255);
      }
#endif

#if 0
    if (mb->motion.prediction_mode == MP2_Intra &&
	d_pichdr->picture_coding_type != MP2_PicType_I)
      {
	DrawLine(d_destimg->AskBitmapY(), lumx   ,lumy,lumx+15,lumy+15, (Pixel)255);
	DrawLine(d_destimg->AskBitmapY(), lumx+15,lumy,lumx   ,lumy+15, (Pixel)255);
      }
#endif

#if 0
    if (mb->motion.prediction_mode == MP2_Frame_FieldBased)
      {
	assert(0);
	DrawLine(img.AskBitmapY(), lumx   ,lumy   ,lumx+15,lumy   , (Pixel)255);
	DrawLine(img.AskBitmapY(), lumx+15,lumy+15,lumx+15,lumy   , (Pixel)255);
	DrawLine(img.AskBitmapY(), lumx   ,lumy+15,lumx+15,lumy+15, (Pixel)255);
	DrawLine(img.AskBitmapY(), lumx   ,lumy   ,lumx   ,lumy+15, (Pixel)255);
      }
#endif

#if 0
    if (mb->motion.prediction_mode == MP2_Frame_FrameBased)
      {
	DrawLine(img.AskBitmapY(), lumx   ,lumy   ,lumx+15,lumy   , (Pixel)255);
	DrawLine(img.AskBitmapY(), lumx+15,lumy+15,lumx+15,lumy   , (Pixel)255);
	DrawLine(img.AskBitmapY(), lumx   ,lumy+15,lumx+15,lumy+15, (Pixel)255);
	DrawLine(img.AskBitmapY(), lumx   ,lumy   ,lumx   ,lumy+15, (Pixel)255);
      }
#endif

#if 0
    DrawPoint(img.AskBitmapY(), lumx   ,lumy,   (Pixel)255);
#endif

#if 0
    if (d_pichdr->picture_coding_type == MP2_PicType_B)
      {
	if (mb->motion.has_mv_forward)
	  {
	    DrawLine(d_destimg->AskBitmapY(),
		     mbx*16+8,mby*16+8,
		     mbx*16+8+ RoundToInt(mb->motion.vector[0][0].h),
		     mby*16+8+ RoundToInt(mb->motion.vector[0][0].v), (Pixel)255);
	  }

	if (mb->motion.has_mv_backward)
	  {
	    DrawLine(d_destimg->AskBitmapY(),
		     mbx*16+8,mby*16+8,
		     mbx*16+8+ RoundToInt(mb->motion.vector[0][1].h),
		     mby*16+8+ RoundToInt(mb->motion.vector[0][1].v), (Pixel)128);
	  }
      }
#endif

    // yp[lumy][lumx]=255;

#if 0
    if (mbx == d_seqhdr->AskWidthInMBs()-1 &&
	mby == d_seqhdr->AskHeightInMBs()-1 &&
	d_have_top && d_have_bottom &&
	d_sink)
      {
	//d_sink->SendImage(*d_destimg);

#if 1
	switch (d_pichdr->picture_coding_type)
	  {
	  case MP2_PicType_I:
	  case MP2_PicType_P:
	    d_sink->SendImage(d_last,0,0,0,0);

	    //SwapImgs(d_last,d_next);
	    break;

	  case MP2_PicType_B:
	    d_sink->SendImage(*d_destimg,0,0,0,0);
	    break;
	  }
#endif

	//sleep(1);
	d_have_top = d_have_bottom = false;
      }
#endif
  }


  void MP2_VideoDecoder::FlushAllPics()
  {
    if (d_data_valid[currentpic][0]) FlushFrame(currentpic);
    if (d_data_valid[nextpic][0])    FlushFrame(nextpic);
  }

}
