/********************************************************************************
    Copyright (C) 2002  Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "scsplitter.hh"
#include <libvideogfx/error.hh>

#include <iostream>
#include <iomanip>

namespace videocoding {
  using namespace std;

  MP2_StartCodeSplitter::MP2_StartCodeSplitter()
    : d_src(NULL),
      d_current(NULL),
      d_next(NULL)
  {
    Reset();

    for (int i=0;i<256;i++)
      d_breakpoint[i] = false;

    d_breakpoint[MP2_StartCode_PictureStart] = true;
  }


  MP2_StartCodeSplitter::~MP2_StartCodeSplitter()
  {
    if (d_current) delete d_current;
    if (d_next)    delete d_next;
  }


  void MP2_StartCodeSplitter::SetSource(MP2_PES_Source& src)
  {
    d_src = &src;
    Reset();
  }


  void MP2_StartCodeSplitter::SetBreakpointAllSC(bool flag)
  {
    for (int i=0;i<256;i++)
      d_breakpoint[i]=flag;
  }


  void MP2_StartCodeSplitter::Reset()
  {
    if (d_current) delete d_current;
    if (d_next)    delete d_next;
    d_current=NULL;
    d_next   =NULL;
    d_curr_pos=0;
  }


#if 1
  void PrintDataHex(uint8* data,uint32 len)
  {
    //cout << "POS: $" << hex << pck->absolute_filepos << dec << endl;

    for (uint32 i=0;i<len;i++)
      {
	if (i%16 == 0)
	  cout << endl << hex << setw(8) << setfill('0') << i << ": ";
      
	cout << hex << setw(2) << setfill('0') << ((int)data[i]) << " ";
      }
    cout << dec << endl;
  }

  void PrintDataBin(uint8* data,uint32 len)
  {
    //cout << "POS: $" << hex << pck->absolute_filepos << dec << endl;

    for (uint32 i=0;i<len;i++)
      {
	if (i%8 == 0)
	  cout << endl << hex << setw(8) << setfill('0') << i << ": ";

	uint8 mask=0x80;
	for (int k=0;k<8;k++)
	  cout << ((data[i]&(mask>>k)) ? '1' : '0');
	cout << ' ';
      }
    cout << dec << endl;
  }
#endif


  void MP2_StartCodeSplitter::AddLookahead()
  {
    /* Append first three bytes of additional data to the end of the first one
       to enable searching for start-codes that begin in the first buffer. */

    Assert(d_next==NULL);

    int n_copy = 3;

    while (n_copy > 0)
      {
	// read ahead one packet

	MP2_SystemPacket* p = d_src->GetNextPacket();
	d_next = dynamic_cast<MP2_PES_Packet*>(p);

	if (p) { Assert(d_next); }
      
	if (d_next)
	  {
	    int len = d_next->data.AskLength();
	    int copylen = ((len>=n_copy) ? n_copy : len);

	    d_current->data.AppendBytes(d_next->data.AskData(),copylen);
	    d_next_overhang = copylen;

	    n_copy -= copylen;

	    if (n_copy > 0)
	      {
		/* Data from this packet is not enough, we have to append data
		   from another packet. */

		delete d_next;
		d_next=NULL;
	      }
	  }
	else
	  {
	    d_next_overhang = 0;
	    return; // no more packet available in input queue
	  }
      }
  }

  MP2_SystemPacket* MP2_StartCodeSplitter::GetNextPacket()
  {
    Assert(d_src);


    // ****************** STEP 1: cut packet into new portions ********************/

    if (d_current==NULL)
      {
	MP2_SystemPacket* p = d_src->GetNextPacket();
	d_current = dynamic_cast<MP2_PES_Packet*>(p);

	if (p) { Assert(d_current); }

	if (d_current==NULL)
	  return NULL;

	AddLookahead();
      }


    uint8* currentbuf = d_current->data.AskData();

    MP2_PES_Packet* newpck = new MP2_PES_Packet;
    newpck->Reset();
    newpck->stream_id = d_current->stream_id;
    newpck->input_filepos   = d_current->input_filepos;
    newpck->data_start_offset = d_current->data_start_offset + d_curr_pos;


    /* If this is a picture header, copy the current PTS and DTS time.
     */

    if (d_current->data.AskLength()>=4 && d_current->data.AskData()[3] == 0x00)
      {
	newpck->has_PTS = d_current->has_PTS;
	newpck->has_DTS = d_current->has_DTS;
	newpck->PTS = d_current->PTS;
	newpck->DTS = d_current->DTS;

	// Remove current PTS/DTS so that it is forwarded to new packets only once.

	d_current->has_PTS = false;
	d_current->has_DTS = false;
      }


    /* Copy data up to next breakpoint to new buffer.
     */

    uint32 pos=d_curr_pos+1;   /* Search-position. Do not search first position, as this is the
				  start-code at the beginning of the packet. */
    uint32 lastsearchpos = d_current->data.AskLength()-4;

    for (;; pos++ )
      {
	if (pos > lastsearchpos)
	  {
	    if (d_next)
	      {
		/* We read through the complete current packet, but found no start-code.
		   Thus, copy all of the current packet (excluding the last overhang-bytes)
		   to the output and continue with the next code.
		   The next-packet will be filled with a new packet if there is one.
		*/

		uint8* copystart = &currentbuf[d_curr_pos];
		uint32 copylen = d_current->data.AskLength()-d_curr_pos-d_next_overhang;

		newpck->data.AppendBytes(copystart,copylen);
              
		// Read in new packet.
              
		delete d_current;
		d_current=d_next;
		Assert(d_current);

		currentbuf = d_current->data.AskData();

		d_next = NULL;
		AddLookahead();
              
		// Reset pointers.
              
		d_curr_pos=0;
		pos = 0;
		lastsearchpos = d_current->data.AskLength()-4;
	      }
	    else
	      {
		// We are at the very end.

		Assert(d_next==NULL);

		uint8* copystart = &currentbuf[d_curr_pos];
		uint32 copylen = d_current->data.AskLength()-d_curr_pos;
		newpck->data.AppendBytes(copystart,copylen);

		delete d_current;
		d_current=NULL;

		return newpck;
	      }
	  }


	if (currentbuf[pos  ] == 0 &&
	    currentbuf[pos+1] == 0 &&
	    currentbuf[pos+2] == 1 &&
	    d_breakpoint[currentbuf[pos+3]])
	  {
	    newpck->data.AppendBytes( &currentbuf[d_curr_pos] , pos-d_curr_pos );
	    d_curr_pos = pos;
	    return newpck;
	  }
      }

    Assert(0);
    return NULL;
  }


  bool MP2_StartCodeSplitter::IsEOF() const
  {
    if (d_current) return false;
    Assert(d_src);
    return d_src->IsEOF();
  }

}
